/* Copyright (c) 1996 University of Utah.  All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for NON-COMMERCIAL purposes and without fee is
 * hereby granted provided that this copyright notice appears in all
 * copies.
 *
 * See http://www.cs.utah.edu/~hamlet for more educational resources.
 *
 * The University of Utah MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT
 * THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT. The University of Utah SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package BlockStacker;

import java.awt.*;

import BlockStacker.*;

//*************************************************************************
// WorkArea class - The area that the blocks are stacked in
//*************************************************************************

class WorkArea extends Canvas {

    public final static Color tableColor = Color.white;
    public final static Color borderColor = Color.black;
    public final static Color blockOKColor = Color.green;
    public final static Color blockBadColor = Color.red;

    private final static int leftWidth = 10;
    private final static int startBlocks = 10;

    private int blockWidth;        // Width of block and table
    private int minBlockWidth;     // Minimum allowable block width
    private int blockHeight;       // Height of block
    private int tableHeight;       // Height of table
    private int maximumBlocks;     // Maximum number of blocks ever
    private int maxBlocks;         // Maximum number of blocks
    private int height;            // Height of display area

    private Block[] theBlocks;     // Blocks in stacking order
    private int topBlock;          // Index of top block

    private int oldBlockWidth;     // Block width when resizing started
    public int blockIndex;         // Index of block where mouse went down
    boolean inTable;               // True if mouse went down in table
    private int mouseX;            // X-coordinate of dragging mouse
    private int offset;
    private int extension = 0;
    

    //--------------------------------------------------
    // Constructor
    //--------------------------------------------------

    WorkArea () {
	topBlock = -1;
	maximumBlocks = 100;
	inTable = false;
	blockIndex = -1;
	theBlocks = new Block[maximumBlocks];
    }

    

    //--------------------------------------------------
    // Draw the graphics to a blank screen
    //--------------------------------------------------

    public synchronized void paint (Graphics g) {

	// Draw the table
	
	g.setColor(tableColor);

	g.fillRect(tableX(), tableY(), blockWidth-1, tableHeight-1);


	// Outline the table and legs
	
	g.setColor(borderColor);
	
	g.drawRect(tableX(), tableY(), blockWidth-1, tableHeight-1);

	g.drawRect(legX(1), legY(), legWidth(), legHeight()-1);

	g.drawRect(legX(2), legY(), legWidth(), legHeight()-1);


	// Draw the legs

	g.setColor(tableColor);

	g.fillRect(legX(1)+1, legY()-1, legWidth()-1, legHeight());

	g.fillRect(legX(2)+1, legY()-1, legWidth()-1, legHeight());


	// Label the table

	g.setColor(borderColor);

	String label = "Length: " + Integer.toString(blockWidth) + " pixels";
	FontMetrics fm = getFontMetrics(getFont());
	int labelWidth = fm.stringWidth("Length: ");
	g.drawString(label,
		     tableX() + blockWidth/2 - labelWidth,
		     tableY() + 2*fm.getLeading() + fm.getAscent());


        // Draw the extension

	drawExtension(g);


	// Draw the blocks

	int x = tableX();
	int y = tableY();

	for (int i = 0; i <= topBlock; i++) {
	    x = x + theBlocks[i].offset;
	    y = y - blockHeight + 1;
	    theBlocks[i].draw(g, x, y, blockWidth, blockHeight);
	}
	
    }
    
    
    public void drawExtension (Graphics g) {
	String pixels = " pixels";
	if (extension == 1) pixels = " pixel";
	String label = "Overhang: " + Integer.toString(extension) + pixels;
	FontMetrics fm = getFontMetrics(getFont());
        int labelWidth = fm.stringWidth("Overhang: ");
	g.setColor(tableColor);
	g.fillRect(tableX() + blockWidth/2,
 	           tableY() + tableHeight - 3*fm.getLeading()
		            - fm.getDescent() - fm.getAscent(),
		   blockWidth/2 - 1,
		   tableHeight/2 - 1);
	g.setColor(borderColor);
        g.drawString(label,
                     tableX() + blockWidth/2 - labelWidth,
                     tableY() + tableHeight - 2*fm.getLeading()
		                            - fm.getDescent());
    }



    //--------------------------------------------------
    // Computes coordinates for the table
    //--------------------------------------------------

    private int tableX () {
	return(leftWidth);
    }

    private int tableY () {
	return(height - legHeight() - tableHeight);
    }

    private int legWidth () {
	return(blockHeight);
    }

    private int legHeight () {
	return(2*blockHeight);
    }

    private int legY () {
	return(tableY() + tableHeight);
    }

    private int legX (int leg) {
	if (leg == 1)
	    return(tableX() + legWidth());
	else
	    return(tableX() + blockWidth - 2*legWidth());
    }
    


    //--------------------------------------------------
    // Reports whether a coordinate is inside the table
    //--------------------------------------------------

    public boolean insideTable (int x, int y) {
	return((tableX() <= x &&
		tableY() <= y &&
		x < tableX()+blockWidth &&
		y < tableY()+tableHeight)
	       ||
	       (legX(1) <= x &&
		legY() <= y &&
		x < legX(1)+legWidth() &&
		y < legY()+legHeight())
	       ||
	       (legX(2) <= x &&
		legY() <= y &&
		x < legX(2)+legWidth() &&
		y < legY()+legHeight()));
    }
    

    //--------------------------------------------------
    // Put one block on the table, extended halfway.
    //--------------------------------------------------

    public synchronized void oneHalfway () {
	for (int i=0; i<=topBlock; i++)
	    theBlocks[i] = null;
	topBlock = 0;
	theBlocks[0] = new Block();
	theBlocks[0].offset = blockWidth / 2;
	Graphics g = getGraphics();
	update(g);
	rebalance(g);
	setMenus();
    }
	
    
    //-----------------------------------------------------
    // Put two blocks, each extended by one-third.
    //-----------------------------------------------------

    public synchronized void twoThirds () {
	for (int i = 0; i <= topBlock; i++)
	    theBlocks[i] = null;
	topBlock = 1;
	theBlocks[0] = new Block();
	theBlocks[1] = new Block();
	theBlocks[0].offset = blockWidth / 3;
	theBlocks[1].offset = blockWidth - 2 * (blockWidth/3);
	Graphics g = getGraphics();
	update(g);
	rebalance(g);
	setMenus();
    }


    // Do a custom stack to produce an image for a paper I've written.

    public synchronized void custom () {
      blockWidth = 240;
      for (int i = 0; i <= topBlock; i++)
	theBlocks[i] = null;
      topBlock = 5;
      theBlocks[0] = new Block();
      theBlocks[1] = new Block();
      theBlocks[2] = new Block();
      theBlocks[3] = new Block();
      theBlocks[4] = new Block();
      theBlocks[5] = new Block();
      theBlocks[0].offset = blockWidth / 12;
      theBlocks[1].offset = blockWidth / 10;
      theBlocks[2].offset = blockWidth / 8;
      theBlocks[3].offset = blockWidth / 6;
      theBlocks[4].offset = blockWidth / 4;
      theBlocks[5].offset = blockWidth / 2;
      Graphics g = getGraphics();
      update(g);
      rebalance(g);
      setMenus();
    }


    //--------------------------------------------------
    // Reset to an empty table
    //--------------------------------------------------

    public synchronized void reset () {
	if (topBlock < 0) return;
	eraseBlock(0);
	for (int i=0; i<=topBlock; i++)
	    theBlocks[i] = null;
	topBlock = -1;
	rebalance(getGraphics());
	setMenus();
    }
    


    //--------------------------------------------------
    // Set the state of the menu items.
    //--------------------------------------------------

    public void setMenus () {
	BlockStacker parent = (BlockStacker) (getParent());
	parent.full(topBlock+1 >= maxBlocks);
	parent.empty(topBlock < 0);
    }


    //--------------------------------------------------
    // Remove the top block
    //--------------------------------------------------

    public synchronized void remBlock () {
	if (topBlock < 0) return;
	eraseBlock(topBlock);
	theBlocks[topBlock] = null;
	topBlock--;
	rebalance(getGraphics());
	setMenus();
    }


    //----------------------------------------------------
    // Add a block on top of the stack
    //----------------------------------------------------

    public synchronized void addBlock () {
	if (topBlock+1 >= maxBlocks) return;
	topBlock = topBlock + 1;
	theBlocks[topBlock] = new Block();
	Graphics g = getGraphics();
	theBlocks[topBlock].draw(g, xcoord(topBlock), ycoord(topBlock),
				 blockWidth, blockHeight);
	rebalance(g);
	setMenus();
    }



    //----------------------------------------------------
    // Add a block to the bottom of the stack
    //----------------------------------------------------

    public synchronized void addBlockBottom () {
	if (topBlock+1 >= maxBlocks) return;
	if (topBlock < 0) {
	    addBlock();
	    return;
	}
	Graphics g = getGraphics();
	g.copyArea(0, ycoord(topBlock),
		   size().width, ycoord(-1)-ycoord(topBlock),
		   0, -blockHeight+1);
	g.clearRect(0, ycoord(0), size().width, blockHeight);
	g.drawLine(tableX(), tableY(), tableX()+blockWidth-1, tableY());
	topBlock = topBlock + 1;
	for (int i = topBlock-1; i >= 0; i--)
	    theBlocks[i+1] = theBlocks[i];
	theBlocks[0] = new Block();
	theBlocks[0].draw(g, xcoord(0), ycoord(0), blockWidth, blockHeight);
	g.drawLine(xcoord(1), ycoord(0), xcoord(1)+blockWidth-1, ycoord(0));
	setMenus();
    }



    //--------------------------------------------------
    // Erase the block whose "index" is given.
    //--------------------------------------------------

    public synchronized void eraseBlock (int index) {
	Graphics g = getGraphics();
	int x = xcoord(index) - theBlocks[index].offset;
	int y = ycoord(index-1);
	int offset = theBlocks[index].offset;
	g.clearRect(0, 0, size().width, y+1);
	g.drawLine(x, y, x+blockWidth-1, y);
    }


    //--------------------------------------------------
    // Return x- and y-coordinates of UL corner of block
    // whose "index" is given.
    //--------------------------------------------------

    public synchronized int ycoord (int index) {
	if (index < 0) return(tableY());
	return(tableY() - (index+1)*(blockHeight-1));
    }
    
    public synchronized int xcoord (int index) {
	int x = tableX();
	for (int i = 0; i <= index; i++)
	    x += theBlocks[i].offset;
	return(x);
    }



    //------------------------------------------------------
    // Updates display according to change since last update
    //------------------------------------------------------

    public synchronized void update (Graphics g) {


	// In this case, the mouse is being dragged inside a block

	if (blockIndex >= 0) {
	    
	    // Obtain the amount by which the mouse has moved
	    // since the last update, and reset the offset of
	    // the block

	    int deltaX = offset - theBlocks[blockIndex].offset;
	    theBlocks[blockIndex].offset = offset;
	    

	    // Obtain the y-coordinate of the lower edge of the
	    // block, the x-coordinate of the left edge of the
	    // block, and y-coordinate of the top edge of the
	    // top block

	    int y = ycoord(blockIndex-1);
	    int x = xcoord(blockIndex);
	    int ytop = ycoord(topBlock);


	    // Move the block and everything above it by deltaX

	    g.copyArea(0,
		       ytop,
		       size().width,
		       y-ytop+1,
		       deltaX,
		       0);


	    // We turn our attention to the line at the bottom of
	    // the block, which is shared with the surface below.
	    // We obtain the x-coodinate of the left and right ends
	    // of this line.

	    int leftX = Math.min(x, x-offset);
	    int rightX = blockWidth-1 + Math.max(x, x-offset);


	    // When we moved the block and everything above it, it
	    // messed up the shared line.  Fix it.

	    if (deltaX > 0) {
		g.setColor(getBackground());
		g.drawLine(rightX+1, y, rightX+deltaX, y);
		g.setColor(borderColor);
		g.drawLine(leftX, y, leftX+deltaX, y);
	    }

	    if (deltaX < 0) {
		g.setColor(getBackground());
		g.drawLine(leftX+deltaX, y, leftX-1, y);
		g.setColor(borderColor);
		g.drawLine(rightX+deltaX, y, rightX, y);
	    }


	    // Update the display of the block's offset, recompute
	    // the maximum extension, and change the colors of
	    // the supporting blocks if necessary.

	    theBlocks[blockIndex].color = null;
	    rebalance(g);
	    

	    // Continue cleaning up after the copy.  We clear the
	    // rectangle of width deltaX across which the area was
	    // copied, and set a clipping rectangle in preparation
	    // for redisplaying this region.

	    if (deltaX > 0) {
		g.clearRect(0,
			    ycoord(topBlock),
			    deltaX,
			    ycoord(blockIndex-1)-ycoord(topBlock)+1);

		g.clipRect(0,
			   ycoord(topBlock),
			   deltaX,
			   ycoord(blockIndex-1)-ycoord(topBlock)+1);
	    }			   

	    else if (deltaX < 0) {
		g.clearRect(size().width+deltaX,
			    ycoord(topBlock),
			    -deltaX,
			    ycoord(blockIndex-1)-ycoord(topBlock)+1);

		g.clipRect(size().width+deltaX,
			   ycoord(topBlock),
			   -deltaX,
			   ycoord(blockIndex-1)-ycoord(topBlock)+1);
	    }
	    

	    // We redisplay the blocks and the line of the surface
	    // below within the clipping region.

	    if (deltaX != 0) {

		g.drawLine(x-offset,
			   y,
			   x-offset+blockWidth-1,
			   y);
	    
		for (int i = blockIndex; i <= topBlock; i++)
		    theBlocks[i].draw(g,
				      xcoord(i),
				      ycoord(i),
				      blockWidth,
				      blockHeight);

	    }

	}


	// In this case the mouse is being dragged inside the table

	else if (inTable) {

	    // Get the width and height of the region that is actually
	    // being used

	    int width = size().width;
	    int height = size().height-ycoord(topBlock);

	    // Compute the extension and colors of the blocks, but
	    // don't redisplay the colors
	    
	    rebalance(null);

	    // Create an off-screen image with a new version of the
	    // block stack

	    Image image = createImage(width, height);
	    Graphics imageG = image.getGraphics();
	    imageG.setFont(g.getFont());
	    imageG.setColor(getBackground());
	    imageG.fillRect(0, 0, width, height);
	    imageG.translate(0, -ycoord(topBlock));
	    paint(imageG);

	    // Render the image in the proper place

	    g.drawImage(image, 0, ycoord(topBlock), getBackground(), null);
	    return;

	}

	// Otherwise we just clear and paint

	else {
	    g.clearRect(0, 0, size().width, size().height);
	    paint(g);
	}

    }



    //--------------------------------------------------
    // Handles mouseDown events
    //--------------------------------------------------

    public synchronized boolean mouseDown (Event evt, int x, int y) {

	// Remember where the event happened

	mouseX = x;


	// In this case we are inside a table

	if (insideTable(x,y)) {
	    inTable = true;
	    blockIndex = -1;
	    oldBlockWidth = blockWidth;
	    for (int i = 0; i <= topBlock; i++)
		theBlocks[i].oldOffset = theBlocks[i].offset;
	}


	// In this case we are inside a block

	else if (tableY()-y > 0) {
	    inTable = false;
	    blockIndex = (tableY()-y)/(blockHeight-1);
	    if (blockIndex > topBlock) {
		blockIndex = -1;
		return(true);
	    }
	    int leftX = xcoord(blockIndex);
	    if (x < leftX || leftX+blockWidth-1 < x) {
		blockIndex = -1;
		return(true);
	    }
	    offset = theBlocks[blockIndex].offset;
	}
	

	return true;
    }

    
    
    //--------------------------------------------------
    // Handles mouseDrag events
    //--------------------------------------------------

    public synchronized boolean mouseDrag (Event evt, int x, int y) {

	int currentX;
	int deltaX;

	// If the mouse is in the table, resize the table and
	// the blocks it supports.

	if (inTable) {

	    // Find the new width of table and blocks

	    blockWidth = blockWidth + (x - mouseX);
	    if (blockWidth < minBlockWidth) blockWidth = minBlockWidth;
	    mouseX = x;

	    // Detemine the percentange by which the width has changed

	    float pct = ((float) blockWidth) / oldBlockWidth;

	    // Resize each of the blocks

	    for (int i=0; i <= topBlock; i++)
		theBlocks[i].offset = Math.round(pct *
						 theBlocks[i].oldOffset);

	}


	// If the mouse is in one of the blocks, move it and
	// all the blocks it supports.

	if (blockIndex >= 0) {

	    // Determine the number of pixels the mouse has moved
	
	    deltaX = x - mouseX;
	    mouseX = x;
	

	    // Calculate how far over the edge the current block is

	    offset = offset + deltaX;

	    // Don't let the block lose contact with the one below it

	    if (offset >= blockWidth) {
		offset = blockWidth - 1;
	    }
	    else if (offset <= -blockWidth) {
		offset = 1 - blockWidth;
	    }

	}

	repaint();

	return(true);

    }

    
    
    //--------------------------------------------------
    // Handles mouseUp events
    //--------------------------------------------------

    public synchronized boolean mouseUp (Event evt, int x, int y) {

	// Bring everything up to date, clear the index and flag
	
	update(getGraphics());

	blockIndex = -1;
	inTable = false;

	return true;
    }
    


    //--------------------------------------------------
    // Set the color of each block according to whether
    // it is balanced.  Also set the maximum extension.
    //--------------------------------------------------

    public void rebalance (Graphics g) {

	float sum = 0;
	int nblocks = 0;

	for (int i=topBlock; i>=0; i--) {

	    sum = sum + blockWidth/2.0f;
	    nblocks++;

	    if (sum/nblocks >= -theBlocks[i].offset &&
		   sum/nblocks <= blockWidth-theBlocks[i].offset) {
		if (theBlocks[i].color != blockOKColor) {
		    theBlocks[i].color = blockOKColor;
		    if (g != null)
			theBlocks[i].recolor(g, xcoord(i), ycoord(i),
					     blockWidth, blockHeight);
		}
	    }
	    else {
		if (theBlocks[i].color != blockBadColor) {
		    theBlocks[i].color = blockBadColor;
		    if (g != null)
			theBlocks[i].recolor(g, xcoord(i), ycoord(i),
					     blockWidth, blockHeight);
		}
	    }

	    sum = sum + nblocks*theBlocks[i].offset;

	}

	int x = 0;
	extension = 0;
	for (int i=0; i<=topBlock; i++) {
	    x = x + theBlocks[i].offset;
	    if (x > extension) extension = x;
	}

	if (g != null) drawExtension(g);

    }	


    //--------------------------------------------------
    // Preferred initial size
    //--------------------------------------------------

    public Dimension preferredSize () {

        FontMetrics fm = getFontMetrics(getFont());

	blockHeight = fm.getAscent() + fm.getDescent() + 2*fm.getLeading();
	tableHeight = 2*fm.getAscent() + 2*fm.getDescent() + 6*fm.getLeading();
	minBlockWidth = fm.stringWidth("Overhang: 000000 pixels ");
	blockWidth = 2*minBlockWidth/10*10;
	
	return(new Dimension(2*blockWidth + leftWidth,
			     startBlocks*blockHeight +
			       tableHeight+legHeight()));
	
    }


    public void setFont (Font f) {
	System.out.println(f);
	super.setFont(f);
    }


    //--------------------------------------------------
    // Handle resize and reshape requests
    //--------------------------------------------------

    public void resize (int w, int h) {
	adjustHeight(h);
	super.resize(w, h);
    }


    public void resize (Dimension d) {
	adjustHeight(d.height);
	super.resize(d);
    }

    public void reshape (int x, int y, int w, int h) {
	adjustHeight(h);
	super.reshape(x,y,w,h);
    }

    public synchronized void adjustHeight (int h) {
	if (blockHeight == 0) return;
	height = h;
        maxBlocks = tableY()/blockHeight;
	if (maxBlocks > maximumBlocks) maxBlocks = maximumBlocks;
	if (maxBlocks < topBlock+1) maxBlocks = topBlock+1;
	height = Math.max(maxBlocks*blockHeight+tableHeight+legHeight(),
			  height);
	setMenus();
    }    


}
